import cv2
import matplotlib.pyplot as plt
import numpy as np
import sys
import os
from PIL import Image

OUTPUT = os.getcwd() + '/' + 'output'

class FaceDetector():
    
    def __init__(self, file_path):
        # print('face detector init')
        self.file_path = file_path
        if os.path.isdir(OUTPUT) == False:
            os.mkdir(OUTPUT)
            print("Created output folder")


    def get_files(self):
        filenames = os.listdir(self.file_path)
        return filenames


    def crop_faces(self):
        # get files
        # print('start cropping')
        pics = self.get_files()
        # print(pics)
        for i, pic in enumerate(pics):
            images = cv2.imread(self.file_path + '/' + pic)
            cascade = cv2.CascadeClassifier("./haarcascade_frontalface_alt.xml")
            face_list = cascade.detectMultiScale(
            images, scaleFactor=1.1, minNeighbors=1, minSize=(1, 1))
            no = 1
            for rect in face_list:
                x = rect[0]
                y = rect[1]
                width = rect[2]
                height = rect[3]
                dst = images[y: y + height, x: x + width]
                save_path = OUTPUT + '/' + 'output_' + str(i) + '_' +  str(no) + '.png'
                result = cv2.imwrite(save_path, dst)
                no += 1





import os
import glob
from PIL import Image

# ImageResizer
# resize target folder's images
#

class ImageResizer():
    
    def __init__(self, width, height):
        self.width = width
        self.height = height

    def get_size(self, filename):
        st = os.stat(filename)
        return st.st_size

    def resize_image(self, target):
        # get all image files
        files = []
        jpgs = glob.glob(target + '/*.jpg') # jpg
        jepegs = glob.glob(target + '/*.jpeg') # jpeg
        pngs = glob.glob(target + '/*.png') # png
        files.extend(jpgs)
        files.extend(jepegs)
        files.extend(pngs)
        # print(files)

        # resize
        for f in files:
            # print(f)
            file_size = self.get_size(f)
            if file_size > 0:
                img = Image.open(f)
                img_resize = img.resize((self.width, self.height))
                ftitle, fext = os.path.splitext(f)
                # img_resize.save(ftitle + '_resized' + fext)
                img_resize.save(ftitle + fext)





# from image_collector import ImageCollector
# import argparse

# parser = argparse.ArgumentParser(description="image download")
# parser.add_argument("-k", "--keyword", dest="keyword", help="type keyword")
# parser.add_argument("-n", "--number", dest="number", help="number of total images", default=100)
# args = parser.parse_args()

# # image keyword
# keyword = args.keyword
# number = int(args.number)

# ic = ImageCollector(keyword, number)
# ic.get_images()

# from image_resizer import ImageResizer
# import os
# import argparse

# parser = argparse.ArgumentParser(description="image download")
# parser.add_argument("-t", "--target", dest="target", help="type keyword")
# args = parser.parse_args()

# # image keyword
# target = args.target

# ir = ImageResizer()
# cwd = os.getcwd()
# ImageResizer.resize_image(cwd + '/dataset/' + target)

# from image_size_check import ImageSizeChecker

# isc = ImageSizeChecker()

# isc.check_size('test')



import os
import argparse
import numpy as np
import time
from .color import Color
from .const import Const

# main
def main():
    cwd = os.getcwd()

    parser = argparse.ArgumentParser(
        prog='ictrainer',
        usage='train a image classifier for keras',
        description='tool for training an image classifier',
        add_help = True 
        )
    parser.add_argument("--mode", dest="mode", help="choose one mode of 3 modes, train mode, test mode, and collect mode ", required=True)
    parser.add_argument("--classes", dest="classes", help="put your classname", type = str, nargs='*')
    parser.add_argument("--batch", dest="batch", help="batch size", default=15)
    parser.add_argument("--epoch", dest="epoch", help="epoch a big value takes so much time to train", default=30)
    parser.add_argument("--mname", dest="mname", help="your model name", default="myModel_")
    parser.add_argument("--lr", dest="lr", help="learning rate", default=1e-3)
    parser.add_argument("--momentum", dest="momentum", help="momentum: Parameter that accelerates SGD in the relevant direction and dampens oscillations.", default=0.9)
    parser.add_argument("--rotation", dest="rotation", help="image rotation for training", default=20)
    parser.add_argument("--keyword", dest="keyword", help="type keyword")
    parser.add_argument("-n", "--number", dest="number", help="number of total images", default=100)
    parser.add_argument("-t", "--target", dest="target", help="target folder")
    parser.add_argument("-iw", "--image_width", dest="width", help="resized image width", default=320)
    parser.add_argument("-ih", "--image_height", dest="height", help="resized image height", default=180)
    args = parser.parse_args()

    # args
    user_define_classes = args.classes # classes defined by user 
    batch = args.batch # batch size
    epoch = args.epoch # epoch size
    mname = args.mname  # model name
    input_lr = args.lr
    input_momentum = args.momentum
    input_rotation = args.rotation
    mode = args.mode # mode
    keyword = args.keyword
    max_num = int(args.number)
    target = args.target
    resized_width = args.width
    resized_height = args.height
    

    print('mode: ' + mode)
    if mode == 'collect':
        # mode image collect images
        from .image_collector import ImageCollector
        print(Color.CYAN + 'start image collecting mode' + Color.END)
        ic = ImageCollector(keyword, max_num)
        ic.get_images()
        print(Color.CYAN + 'end image collecting mode' + Color.END)
        print(Color.PURPLE + 'check files size' + Color.END)
        from .image_size_check import ImageSizeChecker
        isc = ImageSizeChecker()
        # check dataset folder
        folders = os.listdir('dataset')
        for folder in folders:
            if folder != '.DS_Store':
                isc.check_size('dataset/' + folder)
        print(Color.PURPLE + 'removed 0 size file' + Color.END)

    elif mode == 'resize':
        # resize ToDo
        from .image_resizer import ImageResizer
        print(Color.GREEN + 'start image resize mode' + Color.END)
        ir = ImageResizer(resized_width, resized_height)
        ir.resize_image(cwd + '/dataset/' + target)
        print(Color.GREEN + 'end image resize mode' + Color.END)

    elif mode == 'face':
        # face detector
        from .face_detector import FaceDetector
        print(Color.YELLOW + 'start detecting faces' + Color.END)
        fd = FaceDetector(cwd + '/' + target)
        fd.crop_faces()
        print(Color.YELLOW + 'end detecting faces' + Color.END)
        print(Color.RED + 'Please check output folder!!!' + Color.END)

    
    elif mode == 'train':
        # mode train train images
        from keras import optimizers
        from .ictrainer import ICTrainer
        print(Color.BLUE + 'start training mode' + Color.END)

        # create folder for result
        if not os.path.exists(Const.RESULT_DIR):
            print(Color.YELLOW + 'You do not have result dir. I will create it for you!!!' + Color.END)
            os.mkdir(Const.RESULT_DIR)

        classes = user_define_classes
        nb_classes = len(classes)

        # icTrainer instance
        ict = ICTrainer(classes, nb_classes, batch)

        # data folder path for train & validation
        train_data_dir = cwd + '/dataset/train'
        # validation_data_dir = cwd + '/dataset/val'
        train_data_num = ict.get_number_of_files(train_data_dir)
        # validation_data_num = ict.get_number_of_files(validation_data_dir)
        
        # print('train data num ' + str(train_data_num))
        # print('val data num ' + str(validation_data_num))


        # the amount of images
        # count the number of images
        nb_train_samples = train_data_num * nb_classes
        # nb_validation_samples = validation_data_num * nb_classes

        # start timer
        start = time.time()

        # create model
        vgg_model = ict.vgg_model_maker()

        # freeze layers
        for layer in vgg_model.layers[:15]:
            layer.trainable = False

        # classes
        # optimizer lr(lerning rate) # parameter
        # lr: float >= 0. Learning rate.
        # momentum: float >= 0. Parameter that accelerates SGD 
        # in the relevant direction and dampens oscillations.
        # loss
        # metrics
        # 
        vgg_model.compile(loss='categorical_crossentropy',
            optimizer=optimizers.SGD(lr=input_lr, momentum=input_momentum),metrics=['accuracy'])

        # generator for images
        train_generator, validation_generator = ict.image_generator(input_rotation)

        # Fine-tuning
        history = vgg_model.fit_generator(
            train_generator,
            samples_per_epoch=nb_train_samples,
            nb_epoch=epoch,
            validation_steps=20,
            validation_data=validation_generator
            # nb_val_samples=nb_validation_samples
            )

        # export model
        vgg_model.save_weights(os.path.join(Const.RESULT_DIR, '{model_name}.h5'.format(model_name=mname)))

        process_time = int((time.time() - start) / 60)
        print(Color.BLUE + u'Finished training!!! This process took', process_time, u' minutes' + Color.END)

    else:
        print(Color.RED + 'you have not put mode name' + Color.END)


if __name__ == '__main__':
    main()



import os

class ImageSizeChecker():

    def __init__(self):
        print('check image file size')
    
    # check images under target folder
    # if the size is zero remove the image
    def check_size(self, targetPath):
        files = os.listdir(targetPath)

        for file in files:
            file_size = os.path.getsize(targetPath + '/' + file)
            if file_size == 0:
                # remove file
                os.remove(targetPath + '/' + file)
                print(file + ' removed')





from icrawler.builtin import GoogleImageCrawler
import os

class ImageCollector():

    def __init__(self, search_keyword, search_number):
        self.keyword = search_keyword
        self.number = search_number

    def get_images(self):
        if not os.path.isdir('dataset/' + self.keyword):
            os.makedirs('dataset/' + self.keyword)

        crawler = GoogleImageCrawler(storage = {"root_dir" : 'dataset/' + self.keyword})
        crawler.crawl(keyword = self.keyword, max_num = self.number)


import os
from keras.applications.vgg16 import VGG16
from keras.preprocessing.image import ImageDataGenerator
from keras.models import Sequential, Model
from keras.layers import Input, Activation, Dropout, Flatten, Dense
from keras.preprocessing.image import ImageDataGenerator
import numpy as np
from .color import Color
from .const import Const

cwd = os.getcwd()
train_data_dir = cwd + Const.TRAIN
validation_data_dir = cwd + Const.VAL

class ICTrainer():

    def __init__(self, classes, nb_classes,batch_size):
        self.classes = classes
        self.nb_classes = nb_classes
        self.batch_size = batch_size


    def vgg_model_maker(self):
        """
        In this case, use VGG16 except Fully Connected layer.
        Need to make FC for this model and put together them
        """

        # Load VGG16 not need fc include_top = False
        input_tensor = Input(shape=(Const.IMG_WIDTH, Const.IMG_HEIGHT, 3))
        vgg16 = VGG16(include_top=False, weights='imagenet', input_tensor=input_tensor)

        # create FC layer
        top_model = Sequential()
        top_model.add(Flatten(input_shape=vgg16.output_shape[1:]))
        top_model.add(Dense(256, activation='relu'))
        top_model.add(Dropout(0.5)) # ToDo
        top_model.add(Dense(self.nb_classes, activation='softmax'))

        # VGG16 + FC ---> new model
        model = Model(input=vgg16.input, output=top_model(vgg16.output))
        return model
    

    def image_generator(self, rotation):
        """
        load images and create dataset for training and validation
        """ 

        # training data
        train_datagen = ImageDataGenerator(
            rescale=1.0 / 255,
            zoom_range=0.2,
            rotation_range = rotation,
            horizontal_flip=True,
            vertical_flip=True)

        # validation data
        validation_datagen = ImageDataGenerator(rescale=1.0 / 255)

        train_generator = train_datagen.flow_from_directory(
            train_data_dir,
            target_size=(Const.IMG_WIDTH, Const.IMG_HEIGHT),
            color_mode='rgb',
            classes=self.classes,
            class_mode='categorical',
            batch_size=self.batch_size,
            shuffle=True)

        validation_generator = validation_datagen.flow_from_directory(
            validation_data_dir,
            target_size=(Const.IMG_WIDTH, Const.IMG_HEIGHT),
            color_mode='rgb',
            classes=self.classes,
            class_mode='categorical',
            batch_size=self.batch_size,
            shuffle=True)

        return (train_generator, validation_generator)
    
    def get_number_of_files(self, target_dir):
        dir_list = os.listdir(target_dir)
        sorted_list = sorted(dir_list)
        # print(sorted_list)
        # For Mac Users
        if sorted_list[0] == '.DS_Store':
            target = sorted_list[1]
        else:
            # print('here')
            target = sorted_list[0]
        # print('target :' + target)
        files_list = os.listdir(target_dir + '/' + target)
        # print(files_list)
        
        return len(files_list)


from setuptools import setup

setup(name="ictrainer",
      version="0.2.0",
      description="cli tool for training your own image classifier with one line command!",
      long_description=open('readme.md').read(),
      long_description_content_type="text/markdown",
      url="https://github.com/koji/ictrainer",
      author='koji kanao',
      author_email='koji.kanao@nyu.edu',
      license="MIT",
      packages=["ictrainer"],
      scripts=["bin/ictrainer"],
      install_requires=["docopt==0.6.2", "keras", "numpy", "tensorflow", "icrawler", "Pillow", "opencv-python", "numpy", "matplotlib"])
